<?php
namespace Teknisa\FIN\Modules\ContaPagar\Services\Remessa\Banco;

use Fiscal\Entities\Boletpagar;
use Financeiro\Util\Util;
use Teknisa\Model\Financeiro\Entities\Cnabpag;
use Fiscal\Repositories\Contapagar as rep_Contapagar;
use Geral\Entities\Empresa;
use Geral\Entities\Filial;
use Geral\Entities\Contcorr;
use Geral\Entities\Tipobaixapag;
use Teknisa\Libs\Util\DateUtil;
use Geral\Util\TratamentoDeMensagens;
use Teknisa\FIN\Modules\ContaPagar\Services\Remessa\GeraCnabService;


/**
 * Class GeraCnabCaixaService
 * @package Financeiro\Logic\ContasPagar\Cnab\Remessa
 */
class GeraCnabCaixaService extends GeraCnabService {


    CONST TIPO_INS_JUR_CPF = 1;
    CONST TIPO_INS_JUR_CNPJ = 2;
    CONST TIPO_INS_JUR_OUTROS = 3;
    CONST LOTE_DE_SERVICO_HEADER_ARQUIVO = '0000';
    CONST LOTE_DE_SERVICO_TRAILLER_ARQUIVO = '9999';
    CONST ID_REGISTRO_HEADER_ARQUIVO = '0';
    CONST ID_REGISTRO_HEADER_LOTE  = '1';
    CONST ID_REGISTRO_DETALHE = '3';
    CONST ID_REGISTRO_TRAILLER_ARQUIVO = '9';
    CONST ID_REGISTRO_TRAILLER_LOTE = '5';
    CONST PARAMETRO_DE_TRANSMISSAO = '01';
    CONST AMBIENTE_CLIENTE_TESTE = 'T';
    CONST AMBIENTE_CLIENTE_PRODUCAO = 'P';
    CONST CODIGO_REMESSA = 1;
    CONST CODIGO_RETORNO = 2;
    CONST VERSAO_LAYOUT = '080';
    CONST DENSIDADE_DE_GRAVACAO = '01600';
    CONST TIPO_SERVICO_PAGAMENTO_FORNECEDOR = '20';
    CONST VERSAO_LAYOUT_LOTE = '041';
    CONST TIPO_COMPROMISSO_PAGAMENTO_FORNECEDOR = '01';
    CONST MOEDA = 'BRL';
    CONST CODIGO_CONPENSACAO_CAIXA = '104';

    private $qtdRegiLote = 0; //Quantidade de registro do lote
    private $qtdRegiArq  = 0; //Quantidade de registro do arquivo
    private $vrPagtLote  = 0; //Somatório dos registros com tipo de movimento 000 (inclusão)
    private $num_lote = 0; // Sequencial incrementado em 1 para cada lote no arquivo
    private $nrSeqRegLote = 0; //Sequencial de Registros no lote
    private $dsEndeEF;
    private $muniEF;
    private $cepEF;
    private $estadoEF;
    private $formaPgto;



    /**
     * @param Cnabpag $remessa
     * @throws \Exception
     */
    public function __construct(Cnabpag $remessa){

        try {
            $this->nrorg = $remessa->getNrorg();

            $this->remessa = $remessa;

            // Carrega os atributos Filial e Empresa
            $this->filial = Filial::findOneByCdfilialAndNrorg($remessa->getCdfilial(), $remessa->getNrorg());
            $this->empresa = Empresa::findOneByCdempresaAndNrorg($remessa->getCdempresa(), $remessa->getNrorg());

            // Busca dados referente a filial se foi informada
            if ($remessa->getCdfilial()) {
                if ($this->filial->getIdtpijurfili() == Filial::TIPO_INSCRICAO_FISICA){
                    $this->tipoinscricao = self::TIPO_INS_JUR_CPF;
                } elseif($this->filial->getIdtpijurfili() == Filial::TIPO_INSCRICAO_CGC){
                    $this->tipoinscricao = self::TIPO_INS_JUR_CNPJ;
                }else{
                    $this->tipoinscricao = self::TIPO_INS_JUR_OUTROS;
                }

                $this->inscricaojuridica = Util::removeMascaraInscricao($this->filial->getNrinsjurfili());
                $this->nomeempresa = $this->filial->getNmrazsocfili();
                $this->dsEndeEF = $this->filial->getEndefiliprinc()->getDsendefili();
                $this->muniEF = $this->filial->getEndefiliprinc()->getMunicipio()->getNmmunicipio();
                $this->cepEF = $this->filial->getEndefiliprinc()->getNrcepfili();
                $this->estadoEF = $this->filial->getEndefiliprinc()->getSgestado();

            } // Se nao busca dados da Empresa
            else {
                if ($this->empresa->getIdtpijurempr() == Empresa::TIPO_INSCRICAO_FISICA){
                    $this->tipoinscricao = self::TIPO_INS_JUR_CPF;
                }elseif($this->empresa->getIdtpijurempr() == Empresa::TIPO_INSCRICAO_CGC){
                    $this->tipoinscricao = self::TIPO_INS_JUR_CNPJ;
                }else{
                    $this->tipoinscricao = self::TIPO_INS_JUR_OUTROS;
                }

                $this->inscricaojuridica = Util::removeMascaraInscricao($this->empresa->getNrinsjurempr());
                $this->nomeempresa = $this->empresa->getNmrazsocempr();
                $this->dsEndeEF = $this->empresa->getDsendeempr();
                $this->muniEF = $this->empresa->getMunicipio()->getNmmunicipio();
                $this->cepEF = $this->empresa->getNrcepempr();
                $this->estadoEF = $this->empresa->getSgestado();
            }

            $this->contcorr = Contcorr::findOneByCdreducontAndNrorg($remessa->getCdreducont(), $remessa->getNrorg());
        } catch (\Exception $e) {
            throw $e;
        }

    }


     /**
     * //#################################################################//
      //############### REGISTRO HEADER DE ARQUIVO ######################//
      //#################################################################//
     * @return string
     * @throws \Exception
     */
    public function geraHeaderArquivo()
    {
        try {

            $this->qtdRegiArq++;

            //Código do Banco - Posição: 1 a 3 - Picture: 9(3)
            $linha = self::CODIGO_CONPENSACAO_CAIXA;

            //Código do Lote de serviço - Posição: 4 a 7 - Picture: 9(4)
            $linha.= self::LOTE_DE_SERVICO_HEADER_ARQUIVO;

            //Tipo de registro = Registro Header - Posição: 8 a 8 - Picture: 9(1)
            $linha.= self::ID_REGISTRO_HEADER_ARQUIVO;

            //Brancos - Posição: 9 a 17 - Picture: X(9)
            $linha.= str_repeat(' ',9);

            //Tipo inscricao da Filial - Posição: 18 a 18 - Picture: 9(1)
            $linha.= $this->tipoinscricao;

            //CGC filial debitada  - Posição: 19 a 32 - Picture: 9(14)
            $linha.= str_pad(trim($this->inscricaojuridica),14,'0',STR_PAD_LEFT);

            //Código convênio no Banco - Posição: 33 a 38 - Picture: 9(6)
            $identpag = self::getIdentPag();
            if($identpag){
                $linha.= str_pad(Util::campoCnabValido(substr(trim($identpag),0,6)),6,'0',STR_PAD_LEFT);
            }else{
                $linha.= str_repeat('0',6);
            }

            //Parâmetro de Transmissão - Posição: 39 a 40 - Picture: 9(2)
            $linha.= self::PARAMETRO_DE_TRANSMISSAO;

            //Ambiente Cliente - Posição: 41 a 41 - Picture: X(1) - T = TESTE P = PRODUÇÃO
            if(! $this->contcorr->getIdfascnabpag())
                $linha.= self::AMBIENTE_CLIENTE_PRODUCAO;
            else
                $linha.= $this->contcorr->getIdfascnabpag();

            //Ambiente Caixa - Posição: 42 a 42 - Picture: X(1)
            $linha.= ' ';

            //Origem do Aplicativo - Posição 43 a 45 - Picture: X(3)
            $linha.= str_repeat(' ',3);

            //Número de Versão - Posição: 46 a 49 - Picture: 9(4)
            $linha.= str_repeat('0',4);

            //Filler - Posição: 50 a 52 - Picture: X(3)
            $linha.= str_repeat(' ',3);

            //Número da agência debitada - Posição: 53 a 57 - Picture: 9(05)
            $linha.= str_pad(substr(trim($this->contcorr->getCdagencia()),0,4),5,'0',STR_PAD_LEFT);

            //DV da Agência - Posição: 58 a 58 - Picture: X(1)
            $linha.= substr(str_pad(trim($this->contcorr->getCdagencia()),5, '0',STR_PAD_RIGHT), -1);
            
            //Operação - Posição: 59 a 62 - Picture: X(4)
            $linha.= str_pad(trim($this->contcorr->getCdopercontcorr()),4,'0',STR_PAD_LEFT);
            
            //Nº da C/C Debitada - Posição: 63 a 70 - Picture: 9(12)
            //Obs: A última posição da C/C é reservada para o DAC
            $linha.= str_pad(Util::campoCnabValido(substr(self::getContaCorrenteSemDigito(),0,8)),8,'0',STR_PAD_LEFT);

            //DV da conta - Posição: 71 a 71 - Picture: X(1)
            $linha.= Util::campoCnabValido(self::getDigitoContaCorrente());

            //DAC da agência/Conta debitada - Posição: 72 a 72 - Picture: 9(1)
            $linha.= ' ';

            //Nome da empresa - Posição: 73 a 102 - Picture: X(30)
            $linha.= str_pad(substr(trim(Util::campoCnabValido($this->nomeempresa)),0,30),30,' ',STR_PAD_RIGHT);

            //Nome do Banco - Posição: 103 a 132 - Picture: X(30)
            $linha.= str_pad(substr(Util::campoCnabValido($this->contcorr->getAgencia()->getBanco()->getNmbanco()),
                                    0,30),30,' ',STR_PAD_RIGHT);

            //Brancos - Posição: 133 a 142 - Picture: X(10)
            $linha.= str_repeat(' ',10);

            //Código remessa/Retorno  - Posição: 143 a 143 - Picture: 9(1)
            $linha.= self::CODIGO_REMESSA;

            //Data da geração do arquivo - Posição: 144 a 151 - Picture: 9(8)
            $dataAtual = DateUtil::getDataAtual();
            $linha.= $dataAtual->format('dmY');

            //Hora de geração do arquivo - Posição: 152 a 157 - Picture 9(6)
            $linha.= $dataAtual->format('His');

            // CAMPO: Número Seqüencial
            // POSIÇÃO: 158 a 163
            // PICTURE: 9(6)
            if($this->remessa->getNrseqremessapag())
                $linha.= substr($this->remessa->getNrseqremessapag(),-6);
            else
                $linha.= '000001';

            //Versão Layout do arquivo - Posição: 164 a 166 - Picture: 9(3)
            $linha.= self::VERSAO_LAYOUT;

            //Densidade de gravação do arquivo - Posição: 167 a 171 - Picture: 9(5)
            $linha.= self::DENSIDADE_DE_GRAVACAO;

            //Reservado do Banco - Posição: 172 a 191 - Picture: X(20)
            $linha.= str_repeat(' ',20);

            //Reservado da Empresa - Posição: 192 a 211 - Picture: X(20)
            $linha.= str_repeat(' ',20);

            //Uso exclusivo FEBRABAN - Posição: 212 a 222 - Picture: X(11)
            $linha.= str_repeat(' ',11);

            //Ident. Cobrança - Posição: 223 a 225 - Picture: X(03)
            $linha.= str_repeat(' ',3);

            //Uso exclusivo das VAN - Posição: 226 a 228 - Picture: 9(03)
            $linha.= str_repeat('0',3);

            //Tipo de serviço - Posição: 229 a 230 - Picture: X(02)
            $linha.= str_repeat(' ',2);

            //Ocorrência Cob. Sem papel - Posição: 231 a 240 - Picture: X(10)
            $linha.= str_repeat(' ',10);

            return $linha."\r\n";

        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * //#################################################################//
       //################ REGISTRO HEADER DE LOTE   ######################//
       //#################################################################//
     * @param $cdbanco
     * @param $boleto
     * @throws \Exception
     * @internal param $formaLancamento
     * @return mixed
     */
    public function geraHeaderLote($cdbanco, $boleto = null)
    {
        try{

            $this->qtdRegiLote++;
            $this->qtdRegiArq++;

            //Código do banco - Posição: 1 a 3 - Picture: 9(3)
            $linha = self::CODIGO_CONPENSACAO_CAIXA;

            //Código do Lote (Lote de identificação de pagtos - Posição: 4 a 7 - Picture: 9(4)
            $linha .= str_pad(++$this->num_lote,4,'0',STR_PAD_LEFT);

            //Tipo de registro = registro Header de lote - Posição: 8 a 8 - Picture: 9(1)
            $linha.= self::ID_REGISTRO_HEADER_LOTE;

            //Tipo de operação (Fixo "C" = Crédito) - Posição: 9 a 9 - Picture: X(1)
            $linha.= 'C';

            //Tipo de Pagamento - Posição: 10 a 11 - Picture: 9(2)
            $linha.= self::TIPO_SERVICO_PAGAMENTO_FORNECEDOR;

            //Forma de Pagamento - Posição: 12 a 13 - Picture: 9(2)
            $linha.= $this->formaPgto = self::getFormaPagto($cdbanco,$boleto);

            //Nº da versão do layoute do lote - Posição: 14 a 16 - Picture: 9(3)
            $linha.= self::VERSAO_LAYOUT_LOTE;

            //Filler - Posição: 17 a 17 - Picture: X(1)
            $linha.= ' ';

            //Tipo inscrição filial debitada - Posição: 18 a 18 - Picture: 9(1)
            $linha.= $this->tipoinscricao;

            //CGC filial debitada  - Posição: 19 a 32 - Picture: 9(14)
            $linha.= str_pad(trim($this->inscricaojuridica),14,'0',STR_PAD_LEFT);

            //Código do Convênio - Posição: 33 a 38 - Picture: 9(6)
            $identpag  = self::getIdentPag();
            if($identpag)
                $linha.= str_pad(Util::campoCnabValido(substr(trim($identpag),0,6)),6,'0');
            else
                $linha.= str_repeat('0',6);

            //Tipo de Compromisso - Posição: 39 a 40 - Picture: 9(2)
            $linha.= self::TIPO_COMPROMISSO_PAGAMENTO_FORNECEDOR;

            //Código de Compromisso - Posição: 41 a 44 - Picture: 9(4)
            $linha.= '0001';

            //Parâmetro de transmissão - Posição: 45 a 46 - Picture: X(2)
            $linha.= '01';

            //Parâmetro de transmissão - Posição: 47 a 52 - Picture: X(6)
            $linha.= str_repeat(' ', 6);

            //Número da agência debitada - Posição: 53 a 57 - Picture: 9(05)
            $linha.= str_pad(substr(trim($this->contcorr->getCdagencia()), 0, 4), 5,'0',STR_PAD_LEFT);

            //DV da Agência - Posição: 58 a 58 - Picture: X(1)
            $linha.= substr(str_pad(trim($this->contcorr->getCdagencia()),5, '0',STR_PAD_RIGHT), -1);
            
            //Operação - Posição: 59 a 62 - Picture: X(4)
            $linha.= str_pad(trim($this->contcorr->getCdopercontcorr()),4,'0',STR_PAD_LEFT);

            //Nº da C/C Debitada - Posição: 63 a 70 - Picture: 9(8)
            $linha.= str_pad(Util::campoCnabValido(substr(self::getContaCorrenteSemDigito(),0,8)),8,'0',STR_PAD_LEFT);

            //Brancos - Posição: 71 a 71 - Picture: X(1)
            $linha.= trim(Util::campoCnabValido(substr($this->contcorr->getCdcontcorr(),-1)));

            //DAC da agência/conta debitada - Posição: 72 a 72 - Picture: 9(1)
            $linha.= ' ';

            //Nome da empresa/filial debitada - Posição: 73 a 102 - Picture: X(30)
            $linha.= str_pad(Util::campoCnabValido(substr($this->nomeempresa, 0, 30)), 30,' ',STR_PAD_RIGHT);

            //Finalidade dos pagamentos do lote - Posição: 103 a 142 - Picture: X(40)
            $linha.= str_repeat(' ', 40);

            //Endereço da empresa - Posição: 143 a 172 - Picture: X(30)
            $linha.= str_pad(Util::campoCnabValido(substr($this->dsEndeEF, 0, 30)), 30,' ',STR_PAD_RIGHT);

            //Número do Local - Posição: 173 a 177 - Picture: 9(5)
            $linha.= str_repeat('0', 5);

            //Complemento (Casa, Apto, sala, etc) - Posição: 178 a 192 - Picture: X(15)
            $linha.= str_repeat(' ', 15);

            //Nome da cidade - Posição: 193 a 212 - Picture: X(20)
            $linha.= str_pad(Util::campoCnabValido(substr($this->muniEF, 0, 20)), 20,' ',STR_PAD_RIGHT);

            //CEP - Posição: 213 a 220 - Picture: 9(8)
            $linha.= str_pad(Util::campoCnabValido(substr($this->cepEF, 0, 8)), 8,' ',STR_PAD_RIGHT);

            //Sigla do Estado - Posição: 221 a 222 - Picture: X(2)
            $linha.= str_pad(Util::campoCnabValido($this->estadoEF), 2,' ',STR_PAD_RIGHT);

            //Brancos - Posição: 223 a 230 - Picture: X(8)
            $linha.= str_repeat(' ', 8);

            //Ocorrências p/ retorno: 231 a 240 - Picture: X(10)
            $linha.= str_repeat(' ', 10);

            return $linha."\r\n";

        }catch(\Exception $e){
            throw $e;

        }
    }

    /**
     * @param $cdbanco
     * @param Boletpagar $boleto
     * @return string
     */
    private function getFormaPagto($cdbanco, $boleto){

        $idbaixa = $boleto->getTpbaixapag()->getIdbaixautilp();
        
        if($idbaixa == Tipobaixapag::TIPO_BAIXA_CNAB_BOLETO){
            if($cdbanco == self::CODIGO_CONPENSACAO_CAIXA)
                return '30'; //Liquidação de títulos do próprio banco
            else
                return '31'; //Pagamento de títulos de outros bancos
        }else{
            if($idbaixa == Tipobaixapag::TIPO_BAIXA_CNAB_ORDEM_PAG)
                return '10'; //OP à disposição
            else{
                if($idbaixa == Tipobaixapag::TIPO_BAIXA_PAG_ORDEM_TRIB_COD_BARRAS)
                    return '13';
                else{
                    if($cdbanco == self::CODIGO_CONPENSACAO_CAIXA)
                        return '01'; //Crédito em conta corrente
                    else{
                        if($boleto->getVrfinalbpag() <= $boleto->getTpbaixapag()->getVrmaxdoccnab())
                            return '03'; //DOC
                        else
                            return '41';//TED
                    }
                }
            }
        }
    }

    //#################################################################//
    //##################  Registro Trailler de Lote ###################//
    //#################################################################//
    /**
     * @return string
     */
    public function geraTraillerLote($somavalores = null)
    {
        try{
            $this->qtdRegiLote++; //Qtde de registro do lote
            $this->qtdRegiArq++;//Quantidade de registro do arquivo

            //Código do banco - Posição: 1 a 3 - Picture: 9(3)
            $linha = '104';

            //Código do lote do serviço - Posição: 4 a 7 - Picture: 9(4)
            $linha.= str_pad(($this->num_lote), 4,'0',STR_PAD_LEFT);

            //Tipo de registro = Registro Trailer do lote - Posição: 8 a 8 - Picture: 9(1)
            $linha.= self::ID_REGISTRO_TRAILLER_LOTE;

            //Brancos - Posição: 9 a 17 - Picture: X(9)
            $linha.= str_repeat(' ', 9);

            //Total qtd Registro do lote - Posição: 18 a 23 - Picture: 9(6)
            $linha.= str_pad($this->qtdRegiLote, 6, '0', STR_PAD_LEFT);

            //Valor do pagto - Posição: 24 a 41 - Picture: 9(16)V9(2)
            $linha.= str_pad(str_replace(array('.',','),'',
                    number_format($this->vrPagtLote,2)
                       ),18,'0',STR_PAD_LEFT);

            //Zeros - Posição: 42 a 59 - Picture: 9(18)
            $linha.= str_repeat('0', 18);

            //Brancos - Posição: 60 a 65 - Picture: X(6)
            $linha.= str_repeat('0', 6);

            //Brancos - Posição: 66 a 230 - Picture: X(165)
            $linha.= str_repeat(' ', 165);

            //Ocorrências - Posição: 231 a 240 - Picture: X(10)
            $linha.= str_repeat(' ', 10);

            return $linha."\r\n";

        }
        catch(\Exception $e){
            TratamentoDeMensagens::exibirMensagens($e);
        }
    }

    //#################################################################//
    //##################  Registro Trailler de Arquivo ###################//
    //#################################################################//
    /**
     * @return string
     */
    public function geraTrailerArquivo()
    {

        //Código do banco na compensação - Posição: 1 a 3 - Picture: 9(3)
        $linha = self::CODIGO_CONPENSACAO_CAIXA;

        //Código do lote do serviço - Posição: 4 a 7 - Picture: 9(4)
        $linha.= self::LOTE_DE_SERVICO_TRAILLER_ARQUIVO;

        //Tipo de registro = registro trailer de arquivo - Posição: 8 a 8 - Picture: 9(1)
        $linha.= self::ID_REGISTRO_TRAILLER_ARQUIVO;

        //Brancos - Posição: 9 a 17 - Picture: X(9)
        $linha.= str_repeat(' ', 9);

        //Total Qtde de lotes do arquivo - Posição: 18 a 23 - Picture: 9(6)
        $linha.= str_pad($this->num_lote, 6,'0',STR_PAD_LEFT);

        //Total de Qtde de registro do arquivo - Posição: 24 a 29 - Picture: 9(6)
        $linha.= str_pad(++$this->qtdRegiArq, 6,'0',STR_PAD_LEFT);

        //Brancos - Posição: 30 a 35 - Picture: 9(6)
        $linha.= str_repeat('0', 6);

        //Brancos - Posição: 36 a 240 - Picture: X(205)
        $linha.= str_repeat(' ', 205);

        return $linha."\r\n";
    }

    /**
     * @param $caminhoArquivo
     * @param $boletos
     * @param bool $autorizaAgendamento - Parámetro utilizando quando o banco for Bradesco
     * @throws \Exception
     * @internal param $remessa
     * @return mixed
     */
    public function geraArquivo($caminhoArquivo, $boletos, $autorizaAgendamento = false, $idopecnabpag = null)
    {
        try{
            $file = fopen($caminhoArquivo.'.txt', "a+");
            $headerarquivo = self::geraHeaderArquivo();
            fprintf($file, "%s",$headerarquivo);

            // Agrupar boletos por tipo de baixa e por banco favorecido
            //$boletosAgrup = self::agrupaBoletos($boletos, $file);
            $qtdBoletos = count($boletos);

            $headerLoteGerado = false;


            //Percorre as formas de pagamento
            /** @var $boleto Boletpagar  */
            foreach($boletos as $boleto){

                /** @var Boletpagar $boleto */
                $boleto = Boletpagar::findOneByNrorgAndCdfilialAndNrlanbolpag($boleto['NRORG'], $boleto['CDFILIAL'], $boleto['NRLANBOLPAG']);

                $banco = self::getBanco($boleto, $boleto->getTpbaixapag()->getIdbaixautilp());
                $formaPgto = self::getFormaPagto($banco,$boleto);
                if(($formaPgto != '30') and ($formaPgto != '31') and ($formaPgto != '13')){
                    if(! $headerLoteGerado){
                        fprintf($file, "%s",self::geraHeaderLote($banco, $boleto));
                        $headerLoteGerado = true;
                    }
                    fprintf($file, "%s",self::geraSegmentoAJ($boleto,$banco));
                    $qtdBoletos--;

                }
            }

            if($headerLoteGerado)
                fprintf($file, "%s",self::geraTraillerLote());

            //Concessionária
            if($qtdBoletos > 0){

                $headerLoteGerado = false;
                $this->qtdRegiLote = 0; //Quantidade de registro do lote
                $this->vrPagtLote = 0; //Somatório dos registros com tipo de movimento 000 (inclusão)
                $this->nrSeqRegLote = 0;

                foreach($boletos as $boleto){

                    /** @var Boletpagar $boleto */
                    $boleto = Boletpagar::findOneByNrorgAndCdfilialAndNrlanbolpag($boleto['NRORG'], $boleto['CDFILIAL'], $boleto['NRLANBOLPAG']);

                    $banco = self::getBanco($boleto, $boleto->getTpbaixapag()->getIdbaixautilp());
                    $formaPgto = self::getFormaPagto($banco,$boleto);
                    if(($formaPgto == '13')){
                        if(! $headerLoteGerado){
                            fprintf($file, "%s",self::geraHeaderLote($banco, $boleto));
                            $headerLoteGerado = true;
                        }
                        fprintf($file, "%s",self::geraSegmentoAJ($boleto,$banco));
                        $qtdBoletos--;

                    }
                }
                if($headerLoteGerado)
                    fprintf($file, "%s",self::geraTraillerLote());

            }

            //Boleto Bancário
            if($qtdBoletos > 0){

                $headerLoteGerado = false;
                $this->qtdRegiLote = 0; //Quantidade de registro do lote
                $this->vrPagtLote = 0; //Somatório dos registros com tipo de movimento 000 (inclusão)
                $this->nrSeqRegLote = 0;

                foreach($boletos as $boleto){

                    /** @var Boletpagar $boleto */
                    $boleto = Boletpagar::findOneByNrorgAndCdfilialAndNrlanbolpag($boleto['NRORG'], $boleto['CDFILIAL'], $boleto['NRLANBOLPAG']);

                    $banco = self::getBanco($boleto, $boleto->getTpbaixapag()->getIdbaixautilp());
                    $formaPgto = self::getFormaPagto($banco,$boleto);
                    if(($formaPgto == '30') or ($formaPgto == '31')){
                        if(! $headerLoteGerado){
                            fprintf($file, "%s",self::geraHeaderLote($banco, $boleto));
                            $headerLoteGerado = true;
                        }
                        fprintf($file, "%s",self::geraSegmentoAJ($boleto,$banco));
                        $qtdBoletos--;

                    }
                }
                if($headerLoteGerado)
                    fprintf($file, "%s",self::geraTraillerLote());
            }






            fprintf($file, "%s",self::geraTrailerArquivo());
            fclose($file);
        }
        catch (\Exception $e){
            fclose($file);
            throw $e;
        }
    }

    /**
     * @param Boletpagar $boleto
     * @param $cdbanco
     * @return string
     */
    private function geraSegmentoAJ(Boletpagar $boleto, $cdbanco){

        $this->qtdRegiLote++;//Qtde de registro do lote
        $this->qtdRegiArq++; //Quantidade de registro do arquivo

        //Código do banco - Posição: 1 a 3 - Picture: 9(3)
        $linha = self::CODIGO_CONPENSACAO_CAIXA;

        //Código do lote de serviço - Posição: 4 a 7 - Picture: 9(4)
        $linha.= str_pad($this->num_lote,4,'0',STR_PAD_LEFT);

        //Tipo de registro = Regitro de detalhe do lote - Posição: 8 a 8 - Picture: 9(1)
        $linha.= self::ID_REGISTRO_DETALHE;

        //Nº de sequencial registro no lote - Posição: 9 a 13 - Picture: 9(5)
        $linha.= str_pad(++$this->nrSeqRegLote,5,'0',STR_PAD_LEFT);

        //Código do segmento reg. detalhe - Posição: 14 a 14 - Picture: X(1)
        if (($this->formaPgto != '30') and ($this->formaPgto != '31') and ($this->formaPgto != '13')){
           $linha.= 'A'; //Inclusão
        }else{
           if ($this->formaPgto == '13')
              $linha.= 'K';
           else
              $linha.= 'J';
        }

        //Tipo de movimento - Posição: 15 a 17 - Picture: 9(3)
        if ($this->remessa->getIdopecnabpag() == 'E')
           $linha.= '999';
        else{
           $linha.= '000';
           //Somatório dos registros com tipo de movimento = 000 (inclusão)
           $this->vrPagtLote+= round($boleto->getVrfinalbpag(), 2);
        }

        if (($this->formaPgto != '30') and ($this->formaPgto != '31') and ($this->formaPgto <> '13')){
           //Complemento de registro - Posição: 18 a 20 - Picture: 9(3)
           $linha.= '000';

           //Código do Banco - Posição: 21 a 23 - Picture: 9(3)
           $linha.= str_pad(substr($cdbanco, 0, 3), 3,'0',STR_PAD_LEFT);

           if ($cdbanco == self::CODIGO_CONPENSACAO_CAIXA){

               //Complemento de Registro - Posição: 24 a 24 - Picture: 9(1)
               $linha.= '0';

               //Numero da agencia creditada - Posição: 25 a 28 - Picture: 9(4)
               if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_CNAB_ORDEM_PAG)
                   $linha.= str_pad(substr(trim($boleto->getReducont()->getCdagencia()), 0, 4), 4,'0',STR_PAD_LEFT);
               else
                   $linha.= str_pad(substr(trim($boleto->getCdagenbolp()), 0, 4), 4,'0',STR_PAD_LEFT);

               //Complemento de Registro - Posição: 29 a 29 - Picture: 9(1)
               if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_CNAB_ORDEM_PAG)
                   $linha.= substr(trim($boleto->getReducont()->getCdagencia()), -1);
               else
                   $linha.= substr(trim($boleto->getCdagenbolp()), -1,1);

               //Numero da C/C creditada - Posição: 30 a 41 - Picture: 9(12)
               if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_CNAB_ORDEM_PAG){
                   $linha.= str_repeat('0', 5);
                   $digitoConta = ' ';
               }
               else{
                   if(strpos($boleto->getCdccorbolp(),'-') === false){
                       $linha.= str_pad(trim(Util::campoCnabValido(substr($boleto->getCdccorbolp(),0,7))),12,'0',STR_PAD_LEFT);
                       $digitoConta = trim(Util::campoCnabValido(substr($boleto->getCdccorbolp(),-1)));
                   }
                   else{
                       $linha.= substr(str_pad(Util::campoCnabValido(substr($boleto->getCdccorbolp(), 0,
                                                                  strpos($boleto->getCdccorbolp(),'-'))),
                                               12,'0',STR_PAD_LEFT), -12);
                       $digitoConta = trim(Util::campoCnabValido(substr($boleto->getCdccorbolp(), (strpos($boleto->getCdccorbolp(),'-') + 1), 1)));
                   }
               }

              //DV C/C - Posição: 42 a 42 - Picture: X(1)
              $linha.= $digitoConta;

              //DV Agência e C/C - Posição: 43 a 43 - Picture: X(1)
              $linha.= ' ';
           }
           else{
              //Código da agência de destino - Posição: 24 a 28 - Picture: 9(5)
               if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_CNAB_ORDEM_PAG){
                   $linha.= str_repeat('0', 5);
                   $digitoAgencia = ' ';
               }
               else{
                   if(strpos($boleto->getCdagenbolp(),'-') === false){
                       $linha.= str_pad(substr($boleto->getCdagenbolp(), 0, -1), 5,'0',STR_PAD_LEFT);
                       $digitoAgencia = trim(Util::campoCnabValido(substr($boleto->getCdagenbolp(),-1,1)));
                   }
                   else{
                       $linha.= str_pad(substr($boleto->getCdagenbolp(), 0, strpos($boleto->getCdccorbolp(),'-')),
                                        5,'0',STR_PAD_LEFT);
                       $digitoAgencia = trim(Util::campoCnabValido(substr($boleto->getCdagenbolp(), (strpos($boleto->getCdagenbolp(),'-') + 1), 1)));
                   }
               }

              //DV Agência - Posição: 29 a 29 - Picture: 9(1)
              $linha.= $digitoAgencia;

               //Conta Corrente - Posição: 30 a 41 - Picture 9(12)
               if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_CNAB_ORDEM_PAG){
                   $linha.= str_repeat('0', 12);
                   $digitoConta = ' ';
               }
               else{
                   if(strpos($boleto->getCdccorbolp(),'-') === false){
                       $linha.= str_pad(trim(substr($boleto->getCdccorbolp(),0,-1)),12,'0',STR_PAD_LEFT);
                       $digitoConta = trim(Util::campoCnabValido(substr($boleto->getCdccorbolp(),-1)));
                   }
                   else{
                       $linha.= substr(str_pad(Util::campoCnabValido(substr($boleto->getCdccorbolp(), 0,
                               strpos($boleto->getCdccorbolp(),'-'))),
                           12,'0',STR_PAD_LEFT), -12);
                       $digitoConta = trim(substr($boleto->getCdccorbolp(), (strpos($boleto->getCdccorbolp(),'-') + 1), 1));
                   }
               }

              //DV Conta Destino - Posição: 42 a 42 - Picture X(1)
               $linha.= $digitoConta;

              //DV Agência/ Conta Destino - Posição: 43 a 43 - Picture X(1)
              $linha.= ' ';

           }

           //Nome do favorecido - Posição: 44 a 73 - Picture: X(30)
           $linha.= str_pad(Util::campoCnabValido(substr($boleto->getForneced()->getNmfantforn(), 0, 30)), 30,' ',STR_PAD_RIGHT);

           //Número do documento atribuído pela empresa - Posição: 74 a 79 - Picture: 9(6)
           //Sera informado na posição 178 a 217
           $linha.= str_repeat('0', 6);

           //Filler - Posição: 80 a 92 - Picture: X(13)
           $linha.= str_repeat (' ', 13);

           //Tipo de Conta - Posição: 93 a 93 - Picture: X(1)
           if (($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPOBAIXA_TED_CIP) or
              ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_TED_STR)){
               if ($boleto->getCdbancbolp())
                    $linha.= '1';   // conta corrente
               else
                   $linha.= '0';   // conta corrente
           }
           else $linha.= '0';   // sem conta

           //Data do pagamento - Posição: 94 a 101 - Picture: 9(8)
           $linha.= $boleto->getDtvenbolpag()->format('dmY');

           //Tipo da Moeda - Posição: 102 a 104 - Picture: X(3)
           $linha.= self::MOEDA;

           //Zeros - Posição: 105 a 119 - Picture: 9(15)
           $linha.= str_repeat('0', 15);

           //Valor do pagto - Posição: 120 a 134 - Picture: 9(13)V9(2)
           $linha.= str_pad(str_replace(array('.',','),'',
                     number_format($boleto->getVrfinalbpag(),2)
                     ),15,'0',STR_PAD_LEFT);

            //Nosso número - Posição: 135 a 143 - Picture: X(9)
           $linha.= str_repeat('0',9);

           //Brancos - Posição: 144 a 146 - Picture: X(3)
           $linha.= str_repeat(' ', 3);

           //Data efetiva - Posição: 147 a 148 - Picture: 9(2)
           $linha.= '01';

           //Valor efetiva - Posição: 149 a 149 - Picture: 9(1)
           $linha.= 'N';

           //Finalidade Detalhe - Posição: 150 a 150 - Picture: 9(1)
           $linha.= '1';

           //Brancos - Posição: 151 a 152 - Picture: X(2)
           $linha.= substr($boleto->getDtvenbolpag(),0,2);

           //Número Parcelas - Posição: 153 a 154 - Picture: 9(2)
           $linha.= '00';

           //Zeros - Posição: 155 a 162 - Picture: 9(8)
           $linha.= str_repeat('0', 8);

           //Zeros - Posição: 163 a 177 - Picture: X(15)
           $linha.= str_repeat('0', 15);

           //Numero de DOC/OP/CHEQUE - Posição: 178 a 217 - Picture: X(40)
           $linha.= str_pad(trim($boleto->getCdfilial().$boleto->getNrlanbolpag()),40,' ',STR_PAD_RIGHT);

           //Finalidade DOC/OP/CHEQUE - Posição: 218 a 219 - Picture: 9(02)
           if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPOBAIXA_CNAB_DOC)
              $linha.= '01';
           else
              $linha.= '00';

           //Uso FEBRABAN - Posição: 220 a 229 - Picture: X(10)
           $linha.= str_repeat(' ', 10);

           //Aviso ao Favorecido - Posição: 230 a 230 - Picture: 9(1)
           $linha.= '0';
        }
        else{
           if ($this->formaPgto == '13'){
              //Identificador do Produto - Posição: 18 a 18 - Picture: 9(1)
              $linha.= substr($boleto->getCdbarrapag(), 0, 1);

              //Código de Segemnto da Empresa - Posição: 19 a 19 - Picture: 9(1)
              $linha.= substr($boleto->getCdbarrapag(), 1, 1);

              //Identificador de Valor - Posição: 20 a 20 - Picture: 9(1)
              $linha.= substr($boleto->getCdbarrapag(), 2, 1);

              //DV do Código de Barras - Posição: 21 a 21 - Picture: 9(1)
              $linha.= substr($boleto->getCdbarrapag(), 3, 1);

              //Valor do Documento - Posição: 22 a 32 - Picture: 9(11)
              $linha.= substr($boleto->getCdbarrapag(), 4, 11);

              //Identificação da Empresa - Posição: 33 a 36 - Picture: 9(4)
              $linha.= substr($boleto->getCdbarrapag(), 15, 4);

              //Data de Vencimento - Posição: 37 a 44 - Picture: X(8)
              $linha.= $boleto->getDtvenbolpag()->format('dmY');

              //Identificação do Contribuinte - Posição: 45 a 56 - Picture: X(12)
              $linha.= substr($boleto->getCdbarrapag(), 19, 12);

              //Filler - Posição: 57 a 61 - Picture: 9(5)
              $linha.= str_repeat(' ', 5);

              //Filler - Posição: 62 a 73 - Picture: X(12)
              $linha.= str_repeat(' ', 12);

              //Número do Documento na Empresa - Posição: 74 a 79 - Picture: X(6)
              $linha.= str_pad($boleto->getCdfilibaipag(), 6,'0',STR_PAD_LEFT);

              //Filler - Posição: 80 a 93 - Picture: X(14)
              $linha.= str_pad(' ', 14);

              //Data da efetivação  - Posição: 155 a 162 - Picture: 9(8)
              $linha.= str_pad('0', 8);

              //Valor da Efetivação - Posição: 163 a 177 - Picture: 9(15)
              $linha.= str_pad('0', 15);

              //Outras Informações - Posição: 178 a 217 - Picture: X(40)
              $linha.= str_pad(' ', 40);

              //Uso FEBRABAN - Posição: 218 a 229 - Picture: X(12)
              $linha.= str_pad(' ', 12);

              //Aviso ao Favorecido - Posição: 230 a 230 - Picture: X(1)
              $linha.= '0';
           }
           else{
               if ($boleto->getTpbaixapag()->getIdbaixautilp() == Tipobaixapag::TIPO_BAIXA_CNAB_BOLETO){
                   if ($boleto->getCdbarrapag()){
                       $cdbanco = str_pad(substr($boleto->getCdbarrapag(),0,3),3,'0',STR_PAD_LEFT);                                                                                                                                                                                                          $cdmoeda = str_pad(substr($boleto->getCdbarrapag(),3,1),1,'0',STR_PAD_LEFT);
                       $cddigito = str_pad(substr($boleto->getCdbarrapag(),4,1),1,'0',STR_PAD_LEFT);
                       $valor = str_pad(substr($boleto->getCdbarrapag(),5,14),14,'0',STR_PAD_LEFT);
                       $cdcampo = str_pad(substr($boleto->getCdbarrapag(),19,25),25,'0',STR_PAD_LEFT);
                   }
                   else{

                       $linhaBarra = '';
                       $dslinhabarrapag = $boleto->getDslinhbarpag();
                       for($i = 0; $i < strlen($boleto->getDslinhbarpag()); $i++){
                           if(in_array($dslinhabarrapag[$i],array('0','1','2','3','4','5','6','7','8','9')))
                               $linhaBarra.= $dslinhabarrapag[$i];
                       }

                        $cdbanco = str_pad(substr($linhaBarra, 0, 3), 3,'0',STR_PAD_LEFT);
                        $cdmoeda = str_pad(substr($linhaBarra, 3, 1), 1,'0',STR_PAD_LEFT);
                        $cddigito = str_pad(substr($linhaBarra, 32, 1), 1,'0',STR_PAD_LEFT);
                        $valor = str_pad(substr($linhaBarra, 33, 14), 14,'0',STR_PAD_LEFT);
                        $cdcampo = str_pad(substr($linhaBarra, 33, 14), 14,'0',STR_PAD_LEFT);
                   }
                   //Código do Banco(Cd. Barra) - Posição: 18 a 20 - Picture: 9(3)
                   $linha.= $cdbanco;

                   //Código da Moeda(Cd. Barra) - Posição: 21 a 21 - Picture: 9(1)
                   $linha.= $cdmoeda;

                   //Dígito verificador(Cd. Barra) - Posição: 22 a 22 - Picture: 9(1)
                   $linha.= $cddigito;

                   //Fator de Vencimento (Cd. Barra) - Posição: 23 a 26 - Picture: 9(4)
                   $linha.= substr($boleto->getCdbarrapag(),5,4);

                   //Valor do Documento (Cd. Barra) - Posição: 27 a 36 - Picture: 9(10)
                   $linha.= substr($boleto->getCdbarrapag(), 9, 10);

                   //Campo Livre (Cd. Barra) - Posição: 37 a 61 - Picture: 9(25)
                   $linha.= $cdcampo;

                   //Nome do favorecido - Posição: 62 a 91 - Picture: X(30)
                   $linha.= str_pad(Util::campoCnabValido(substr($boleto->getForneced()->getNmfantforn(), 0, 30)), 30,' ',STR_PAD_RIGHT);

                   //Data vencimento - Posição: 92 a 99 - Picture: 9(8)
                   $linha.= $boleto->getDtvenbolpag()->format('dmY');

                   //Valor previsto do pagto - Posição: 100 a 114 - Picture: 9(13)V9(2)
                   $linha.= str_pad(str_replace(array('.',','),'',
                                    number_format($boleto->getVrfinalbpag(),2)
                                     ),15,'0',STR_PAD_LEFT);

                   //Valor do Desconto - Posição: 115 a 129 - Picture: 9(13)V9(2)
                   $linha.= str_repeat('0', 15);

                   //Valor do Acrescimo - Posição: 130 a 144 - Picture: 9(13)V9(2)
                   $linha.= str_repeat('0', 15);

                   //Data do pagamento - Posição: 145 a 152 - Picture: 9(8)
                   $linha.= $boleto->getDtvenbolpag()->format('dmY');

                   //Valor do pagto - Posição: 153 a 167 - Picture: 9(13)V9(2)
                   $linha.= str_pad(str_replace(array('.',','),'',
                                    number_format($boleto->getVrfinalbpag(),2)
                                    ),15,'0',STR_PAD_LEFT);

                   //Zeros - Posição:  168 a 182 - Picture: 9(15)
                   $linha.= str_repeat('0', 15);

                   //Seu número - Posição: 183 a 188 - Picture: X(6)
                   //$linha.= PadR(AllTrim(cdsBoletosCDFILIBPAG.AsString  + cdsBoletosNRLANBOLPAG.AsString), 20);
                   $linha.= str_pad(substr($boleto->getNrlanbolpag(),4 ,6),6,'0', STR_PAD_LEFT );

                   //Filler - Posição: 189 a 202 - Picture: X(14)
                   $linha.= str_repeat(' ', 14);

                   //Brancos - Posição: 203 a 211 - Picture: 9(9)
                   $linha.= str_repeat(' ', 9);

                   //Brancos - Posição: 212 a 222 - Picture: X(11)
                   $linha.= str_repeat(' ', 11);

                   //Brancos - Posição: 223 a 224 - Picture: 9(2)
                   $linha.= '09';
               }

           }

        }
        //Brancos - Posição: 225 a 230 - Picture: X(6)
        $linha.= str_repeat (' ', 6);
        
        //Brancos - Posição: 231 a 240 - Picture: X(10)
        $linha.= str_repeat (' ', 10);
        
        return $linha."\r\n";
    }

}